import torch
from torchvision.transforms import Resize, Compose, ToTensor, Normalize
import torch.nn.functional as F

from scipy import signal
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
import math
from utils import *

def cache_indv_layers(layers_cache_indv, inv_corr):
    bs = inv_corr.shape[0]
    if len(layers_cache_indv)==0:
        for i in range(bs):
            layers_cache_indv.append([])
    for i in range(bs):
        layers_cache_indv[i].append(inv_corr[i:i+1])
    return layers_cache_indv

def cache_indv_layers_io(layers_cache_indv, idx, operation, insert_layers=[], selection=None):
    '''
    idx - which layer to change
    selection - where to change
    '''
    indv_layer = []
    bs = len(layers_cache_indv)
    if type(selection) == type(None):
        selection = range(bs)
    num_changes = len(selection)
    if type(idx) != list:
        idx = [idx, ]*num_changes

    if operation=='pop':
        for i in range(num_changes):
            indv_layer.append(layers_cache_indv[selection[i]].pop(idx[i]))
    if operation=='insert':
        for i in range(num_changes):
            layers_cache_indv[selection[i]].insert(idx[i], insert_layers[i])
    return layers_cache_indv, indv_layer

def create_layer_cache(layers_cache_indv):
    layer_cache = []
    bs = len(layers_cache_indv)
    num_layers = len(layers_cache_indv[0])
    for i in range(num_layers):
        layer = torch.cat([layers_cache_indv[j][i] for j in range(bs)], dim=0)
        layer_cache.append(layer)
    return layer_cache

def gaussian_fn(M, std):
    n = torch.arange(0, M, dtype=torch.float) - (M - 1.0) / 2.0
    sig2 = 2 * std * std
#     import pdb; pdb.set_trace()
    w = torch.exp(-n ** 2 / sig2)
    return w

def gkern(kernlen=256, std=128):
    """Returns a 2D Gaussian kernel array."""
    gkern1d = gaussian_fn(kernlen, std=std)
    gkern2d = gkern1d[:, None]* gkern1d[None, :]
    return gkern2d


def plot_log(i, img, corr, part_sig, layers_cache, current_image, final_composite, num_parts, background=None):
    input_np = img.permute(0,2,3,1).detach().cpu().numpy()
    part_np = part_sig.permute(0,2,3,1).detach().cpu().numpy()
    part_np[:,:,:,:3] = part_np[:,:,:,:3]*part_np[:,:,:,3:]
    current_image_np = current_image[-1].permute(0,2,3,1).detach().cpu().numpy()
    final_composite_np = final_composite.permute(0,2,3,1).detach().cpu().numpy()

    fig, ax = plt.subplots(1, 5, figsize=(20,4))
    for a in ax:
        a.axis('off')

    ax[0].set_title('Input', fontsize=24)
    ax[0].imshow(input_np[0,:,:,:3])
    
    for j in range(num_parts):
        ax[j+1].imshow(part_np[j,:,:,:3]*part_np[j,:,:,3:])
        ax[j+1].set_title('Concept {}'.format(j+1), fontsize=24)
    ax[j+2].imshow(current_image_np[0], vmin=0, vmax=1)
    ax[j+2].set_title('Residual', fontsize=24)
    ax[j+3].imshow(final_composite_np[0,:,:,:3])
    ax[j+3].set_title('Reconstruction', fontsize=24)
    plt.show()
    plt.close()

def norm_ncc2D(image, parts, global_alpha, alpha_A, padding):
    n_channels = parts.shape[1]-1
    part_rgb = parts[:,:n_channels]
    part_alpha = parts[:,n_channels:]
    
    part_filter = part_rgb*(part_alpha)
    corr = F.conv2d(image*global_alpha, part_filter, padding=padding)

    if type(alpha_A) == type(None):
        return corr
    I_alpha_a = image*alpha_A
    I_alpha_a = I_alpha_a.sum(dim=1, keepdims=True).sum(dim=2, keepdims=True).sum(dim=3, keepdims=True)
    I_norm = (image**2).sum(dim=1, keepdims=True).sum(dim=2, keepdims=True).sum(dim=3, keepdims=True)
    
    alpha_a_sigma = (alpha_A**2).sum(dim=1, keepdims=True).sum(dim=2, keepdims=True).sum(dim=3, keepdims=True)
    denom = F.conv2d(global_alpha**2, part_filter**2, padding=padding)#+ 1e-8
    ab_2 =  F.conv2d(2*alpha_A*global_alpha, part_filter, padding=padding)
#     denom = torch.sqrt(denom1 * denom2) + 1e-8
    norm_corr = (corr+I_alpha_a)/torch.sqrt(I_norm*(denom+alpha_a_sigma+ab_2))
    safe_tensor = torch.where(torch.isnan(norm_corr), torch.zeros_like(norm_corr), norm_corr)
    return safe_tensor


def quant_mask(corr, return_values=False):
    corr_unroll = corr.view(corr.size(0), -1) # tensor of shape (batch_size, height * width)
    val, idx = corr_unroll.max(dim = 1, keepdim=True)
    mask = torch.zeros_like(corr_unroll)
    for i in range(mask.shape[0]):
        mask[i, idx[i]] = 1
    mask_reshape = mask.reshape(corr.shape)
    if return_values:
        return mask_reshape, val
    return mask_reshape


def hard_composite(layers, shape=[8,3,64,64], background_tensor=None):
    n = len(layers)
    if n==0:
        return torch.zeros(shape).cuda()
    n_channels = layers[0].shape[1]-1
    alpha = (1 - layers[0][:, n_channels:, :, :])
    rgb = layers[0][:, :n_channels] * layers[0][:, n_channels:, :, :]
    for i in range(1, n):
        rgb = rgb + layers[i][:, :n_channels] * layers[i][:, n_channels:, :, :] * alpha
        alpha = (1-layers[i][:, n_channels:, :, :]) * alpha
    if type(background_tensor)!=type(None):
        if background_tensor.shape[1] ==4:
            background = background_tensor[:,:n_channels] * background_tensor[:,n_channels:]
        rgb = rgb + alpha*background[:,:n_channels]
    return rgb


def load_img(img_path, size=(64, 64), RGB=True):
    transform = Compose([
        ToTensor(),
    ])
    if RGB:
        img_pil =Image.open(img_path).convert('RGB').resize((size[0],size[1]), Image.ANTIALIAS)
    else:
        img_pil =Image.open(img_path).convert('RGBA').resize((size[0],size[1]), Image.ANTIALIAS)
    img_tensor = transform(img_pil)
    return img_tensor

def load_batch():
    files = ['data/pattern1-01.png',
             'data/pattern1-02.png',
             'data/pattern1-03.png',
             'data/pattern1-04.png',
             'data/pattern1-05.png',
            ]
    cache = []
    for file in files:
        cache.append(load_img(file, (64,64)))
    batch = torch.stack(cache, dim=0)
    return batch

def hard_composite_layers(layers):
    n = len(layers)
    if n==0:
        return torch.zeros_like(layers[n - 1][:, :3])
    alpha = (1 - layers[0][:, 3:4, :, :])
    rgb = layers[0][:, :3] * layers[0][:, 3:4, :, :]
    layers_output = []
    layers_output.append(rgb)
    for i in range(1, n):
        rgb = rgb + layers[i][:, :3] * layers[i][:, 3:4, :, :] * alpha
        layers_output.append(layers[i][:, :3] * layers[i][:, 3:4, :, :] * alpha)
        alpha = (1-layers[i][:, 3:4, :, :]) * alpha
    return layers_output